import logging
import time

log = logging.getLogger("udata-hydra")


class Timer:
    """
    A simple Timer class for code execution time measurement as a debug log.

    ```
    timer = Timer("my-timer")
    timer.mark("a-step")
    timer.stop()
    ```
    """

    steps = []

    def __init__(self, name: str) -> None:
        self.name = name
        self.steps.append(time.perf_counter())

    def mark(self, step: str) -> None:
        t_mark = time.perf_counter()
        t_delta = t_mark - self.steps[-1]
        self.steps.append(t_mark)
        log.debug(f"[{self.name}] {step} done in {t_delta:0.4f}s")

    def stop(self) -> None:
        t_delta = time.perf_counter() - self.steps[0]
        log.debug(f"[{self.name}] Total time {t_delta:0.4f}s")



import json

from aiohttp import web
from asyncpg import Record

from udata_hydra.db.resource import Resource
from udata_hydra.db.resource_exception import ResourceException
from udata_hydra.utils import get_request_params


async def get_all_resources_exceptions(request: web.Request) -> web.Response:
    """Endpoint to get all resource exceptions from the DB
    Respond with a 200 status code with a list of resource exceptions
    """
    resource_exceptions: list[Record] = await ResourceException.get_all()
    return web.json_response(resource_exceptions)


async def create_resource_exception(request: web.Request) -> web.Response:
    """Endpoint to receive a resource exception creation event from a source
    Will create a new exception in the DB "resources_exceptions" table.
    Respond with a 201 status code with the created resource exception
    If error, respond with a 400 status code
    """
    try:
        payload = await request.json()
        resource_id: str = payload["resource_id"]
        table_indexes: dict[str, str] = payload["table_indexes"]
        # format should be like this:
        #    {
        #       "column_name": "index_type",
        #       "column_name": "index_type"
        #       ...
        #    },
    except Exception as err:
        raise web.HTTPBadRequest(text=json.dumps(err))

    try:
        resource_exception: Record = await ResourceException.insert(
            resource_id=resource_id,
            table_indexes=table_indexes,
        )
    except ValueError as err:
        raise web.HTTPBadRequest(text=f"Resource exception could not be created: {err}")

    return web.json_response(resource_exception, status=201)


async def delete_resource_exception(request: web.Request) -> web.Response:
    """Endpoint to delete a resource exception from the DB
    Respond with a 204 status code
    If error, respond with a 400 status code
    """
    [resource_id] = get_request_params(request, params_names=["resource_id"])
    resource: Record | None = await Resource.get(resource_id)
    if not resource:
        raise web.HTTPNotFound()

    await ResourceException.delete(resource_id)

    return web.HTTPNoContent()



from typing import Callable

from aiohttp import web

from udata_hydra.routes.checks import create_check, get_all_checks, get_latest_check
from udata_hydra.routes.resources import (
    create_resource,
    delete_resource,
    get_resource,
    get_resource_status,
    update_resource,
)
from udata_hydra.routes.resources_exceptions import (
    create_resource_exception,
    delete_resource_exception,
    get_all_resources_exceptions,
)
from udata_hydra.routes.resources_legacy import (
    create_resource_legacy,
    delete_resource_legacy,
    get_resource_legacy,
    update_resource_legacy,
)
from udata_hydra.routes.status import get_crawler_status, get_health, get_stats, get_worker_status


def generate_routes(
    routes_params: list[tuple[Callable, str, Callable, str | None]],
) -> list[web.RouteDef]:
    """
    Generate an aiohttp routes list of web.RouteDef objects from the given route parameters, with each of them having a variant with a trailing slash and one without, since aiohttp does not handle optional trailing slashes easily.
    Args:
        routes_params:
            A list of tuples, where each tuple contains:
                - method: The HTTP method (e.g., web.get, web.post).
                - path: The route path as a string.
                - handler: The handler function for the route.
                - name: An optional name for the route which can later be used by request.app.router to get the route.
    """
    routes: list[web.RouteDef] = []
    for method, path, handler, name in routes_params:
        routes.append(method(path, handler, name=name))
        if path.endswith("/"):
            routes.append(method(path[:-1], handler))
        else:
            routes.append(method(path + "/", handler))
    return routes


# Define the routes parameters
routes_params = [
    (web.get, "/api/checks/latest", get_latest_check, "get-latest-check"),
    (web.get, "/api/checks/all", get_all_checks, None),
    (web.post, "/api/checks", create_check, None),
    # Routes for resources
    (web.get, "/api/resources/{resource_id}", get_resource, None),
    (web.get, "/api/resources/{resource_id}/status", get_resource_status, None),
    (web.post, "/api/resources", create_resource, None),
    (web.put, "/api/resources/{resource_id}", update_resource, None),
    (web.delete, "/api/resources/{resource_id}", delete_resource, None),
    # Routes for statuses
    (web.get, "/api/status/crawler", get_crawler_status, None),
    (web.get, "/api/status/worker", get_worker_status, None),
    (web.get, "/api/stats", get_stats, None),
    (web.get, "/api/health", get_health, None),
    # Routes for resources exceptions
    (web.get, "/api/resources-exceptions", get_all_resources_exceptions, None),
    (web.post, "/api/resources-exceptions", create_resource_exception, None),
    (
        web.delete,
        "/api/resources-exceptions/{resource_exception_id}",
        delete_resource_exception,
        None,
    ),
]

# Generate the routes
routes: list[web.RouteDef] = generate_routes(routes_params)

# TODO: legacy, to remove
legacy_routes_params = [
    (web.get, "/api/resources", get_resource_legacy, None),
    (web.post, "/api/resource/created", create_resource_legacy, None),
    (web.post, "/api/resource/updated", update_resource_legacy, None),
    (web.post, "/api/resource/deleted", delete_resource_legacy, None),
]
legacy_routes: list[web.RouteDef] = generate_routes(legacy_routes_params)
routes.extend(legacy_routes)



import json


async def detect_tabular_from_headers(check: dict) -> tuple[bool, str]:
    """
    Determine from content-type header if file looks like:
        - a csv
        - a csv.gz (1. is the file's content binary?, 2. does the URL contain "csv.gz"?)
        - a xls(x)
    """
    headers: dict = json.loads(check["headers"] or "{}")

    if any(
        headers.get("content-type", "").lower().startswith(ct)
        for ct in ["application/csv", "text/plain", "text/csv"]
    ):
        return True, "csv"

    if any(
        headers.get("content-type", "").lower().startswith(ct)
        for ct in ["application/octet-stream", "application/x-gzip", "application/gzip"]
    ) and "csv.gz" in check.get("url", ""):
        return True, "csvgz"

    if any(
        headers.get("content-type", "").lower().startswith(ct)
        for ct in [
            "application/vnd.ms-excel",
        ]
    ):
        # and "xls" in check.get("url", "")
        return True, "xls"

    if any(
        headers.get("content-type", "").lower().startswith(ct)
        for ct in [
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        ]
    ):
        # and "xlsx" in check.get("url", "")
        return True, "xlsx"

    return False, "csv"
